# Copyright (C) MindEarth <enrico.ubaldi@mindearth.org> @ Mindearth 2020-2021
# 
# This file is part of mobilkit.
#
# mobilkit is distributed under the MIT license.

'''`loader.py` contains a set of tools to load and prepare the database from raw files.
'''
import pytz

import pandas as pd
import numpy as np
import os
import sys

from datetime import datetime, timedelta
from glob import glob

from dask import dataframe as dd
from dask import bag as db
import pytz

from mobilkit.dask_schemas import eventLineRAW

from mobilkit.dask_schemas import (
    accColName,
    lonColName,
    latColName,
    uidColName,
    utcColName,
    dttColName,
    zidColName,
)

def load_raw_files(pattern, version="hflb",
                   timezone=None, start_date=None, stop_date=None,
                   minAcc=300, sep="\t", file_schema=None, **kwargs):
    '''Function that loads the files and returns the dataframe. 

    Parameters
    ----------
    pattern : str
        The pattern of the raw files with bash syntax. For example: ``'sample_data/20*/part-*.csv.gz'``.
    version : str, optional
        One of `hflb`, `wb` or `csv`, the format in which data are stored.
    timezone : str, optional
        The timezone in pytz syntax (e.g., "Europe/Rome" or "America/Mexico_City") to be
        used to localize the Unix Time Stamp time-stamp in the raw-data. If no timezone is specified (default) it defaults to UTC.
    start_date : str, optional
        The starting date when to consider data in the "yyyy-mm-dd" format. This will be localized in `timezone` if given.
    stop_date : str, optional
        The end day up to which consider data in the "yyyy-mm-dd" format. This will be localized in `timezone` if given. This day will be INCLUDED.
    minAcc : int, optional
        The minimum accuracy for a point to be kept. If accuracy is larger than this the point will be discarded. **NOTE that the accuracy column must be called** ``acc``.
    sep : str, optional
        The delimiter of the fields in the files.
    file_schema : list of tuples
        The schema of the file. By default will be :attr:`dask_schemas.eventLineRAW`.
        If `version=wb` `file_schema` is a dictionary telling how to translate the
        original colums in the mobilkit nomenclature.
        **NOTE that the accuracy column must be called** ``acc``.
    **kwargs
        Will be passed to :attr:`mobilkit.loader.load_raw_files_hflb`
        if `version='hflb'`  otherwise to :attr:`mobilkit.loader.load_raw_files_wb`
        if `version='wb'`.

    Returns
    -------
    df : dask.dataframe
        A representation of the dataframe.
    '''
    assert version in ["hflb","wb", "csv"]
    if version == "hflb":
        return load_raw_files_hflb(pattern=pattern, timezone=timezone,
                         start_date=start_date, stop_date=stop_date,
                         minAcc=minAcc, sep=sep, file_schema=file_schema, **kwargs)
    elif version == "wb":
        return load_raw_files_wb(pattern=pattern, timezone=timezone,
                         start_date=start_date, stop_date=stop_date,
                         minAcc=minAcc, sep=sep, file_schema=file_schema, **kwargs)
    elif version == "csv":
        return load_raw_files_custom(pattern=pattern, timezone=timezone,
                         start_date=start_date, stop_date=stop_date,
                         minAcc=minAcc, sep=sep, file_schema=file_schema, **kwargs)
    else:
        return None

def load_raw_files_custom(pattern, timezone=None,
                          start_date=None, stop_date=None,
                          minAcc=300, sep="\t", file_schema=None, partition_size=None, **kwargs):
    '''Function that loads the files and returns the dask dataframe. Note that this
    function is **lazy** meaning that it only construct the dataframe and does not
    build it (it will be built the first time a query is performed on it).

    Parameters
    ----------
    pattern : str
        The pattern of the raw files with bash syntax. For example: ``'sample_data/20*/part-*.csv.gz'``.
    timezone : str, optional
        The timezone in pytz syntax (e.g., "Europe/Rome" or "America/Mexico_City") to be
        used to localize the Unix Time Stamp time-stamp in the raw-data. If no timezone is specified (default) it defaults to UTC.
    start_date : str, optional
        The starting date when to consider data in the "yyyy-mm-dd" format. This will be localized in `timezone` if given.
    stop_date : str, optional
        The end day up to which consider data in the "yyyy-mm-dd" format. This will be localized in `timezone` if given. This day will be INCLUDED.
    minAcc : int, optional
        The minimum accuracy for a point to be kept. If accuracy is larger than this the point will be discarded. **NOTE that the accuracy column must be called** ``acc``.
    sep : str, optional
        The delimiter of the fields in the files.
    file_schema : list of tuples
        The schema of the file. By default will be :attr:`dask_schemas.eventLineRAW`.
        **NOTE that the accuracy column must be called** ``acc``.

    Returns
    -------
    df : dask.dataframe
        A representation of the dataframe.
    '''

    if file_schema is None:
        file_schema = eventLineRAW
    def tmp_filter(r):
        if len(r) != len(file_schema): return False
        for e in r:
            if len(e) == 0:
                return False
        return True
    sp_df = db.read_text(pattern)\
            .str.strip().str.split(sep)\
            .filter(tmp_filter)\
            .to_dataframe(file_schema)
    if partition_size is not None:
        sp_df = sp_df.repartition(partition_size=partition_size)
    if timezone is None:
        timezone = "UTC"
    selected_tz = pytz.timezone(timezone)
    sp_df = filterStartStopDates(sp_df, start_date, stop_date, selected_tz)
    sp_df = compute_datetime_col(sp_df, timezone)
    if minAcc is not None and minAcc > 0:
        print("Filtering on accuracy <= %d" % minAcc)
        sp_df = sp_df[sp_df[accColName] <= minAcc]
    return sp_df

def load_raw_files_hflb(pattern, timezone=None,
                          start_date=None, stop_date=None,
                          minAcc=300, sep="\t", file_schema=None, partition_size=None):
    '''Function that loads the files and returns the dask dataframe. Note that this
    function is **lazy** meaning that it only construct the dataframe and does not
    build it (it will be built the first time a query is performed on it).

    Parameters
    ----------
    pattern : str
        The pattern of the raw files with bash syntax. For example: ``'sample_data/20*/part-*.csv.gz'``.
    timezone : str, optional
        The timezone in pytz syntax (e.g., "Europe/Rome" or "America/Mexico_City") to be
        used to localize the Unix Time Stamp time-stamp in the raw-data. If no timezone is specified (default) it defaults to UTC.
    start_date : str, optional
        The starting date when to consider data in the "yyyy-mm-dd" format. This will be localized in `timezone` if given.
    stop_date : str, optional
        The end day up to which consider data in the "yyyy-mm-dd" format. This will be localized in `timezone` if given. This day will be INCLUDED.
    minAcc : int, optional
        The minimum accuracy for a point to be kept. If accuracy is larger than this the point will be discarded. **NOTE that the accuracy column must be called** ``acc``.
    sep : str, optional
        The delimiter of the fields in the files.
    file_schema : list of tuples
        The schema of the file. By default will be :attr:`dask_schemas.eventLineRAW`.
        **NOTE that the accuracy column must be called** ``acc``.

    Returns
    -------
    df : dask.dataframe
        A representation of the dataframe.
    '''

    if file_schema is None:
        file_schema = eventLineRAW
    
    def tmp_filter(r):
        if len(r) != len(file_schema): return False
        for e in r:
            if len(e) == 0:
                return False
        return True


    sp_df = db.read_text(pattern)\
            .str.strip().str.split(sep)\
            .filter(tmp_filter)\
            .to_dataframe(file_schema)

    if partition_size is not None:
        sp_df = sp_df.repartition(partition_size=partition_size)
    
    if timezone is None:
        timezone = "UTC"
    
    
    selected_tz = pytz.timezone(timezone)

    sp_df = filterStartStopDates(sp_df, start_date, stop_date, selected_tz)

    sp_df = compute_datetime_col(sp_df, timezone)
    
    if minAcc is not None and minAcc > 0:
        print("Filtering on accuracy <= %d" % minAcc)
        sp_df = sp_df[sp_df[accColName] <= minAcc]

    return sp_df


def load_raw_files_wb(pattern, timezone=None, header=False,
                          start_date=None, stop_date=None,
                          minAcc=300, sep="\t", file_schema=None, **kwargs):
    '''Function that loads the files and returns the dask dataframe. Note that this
    function is **lazy** meaning that it only construct the dataframe and does not
    build it (it will be built the first time a query is performed on it).

    Parameters
    ----------
    pattern : str
        The pattern of the raw files with bash syntax. For example: ``'sample_data/20*/part-*.csv.gz'``.
    timezone : str, optional
        The timezone in pytz syntax (e.g., "Europe/Rome" or "America/Mexico_City") to be
        used to localize the Unix Time Stamp time-stamp in the raw-data. If no timezone is specified (default) it defaults to UTC.
    start_date : str, optional
        The starting date when to consider data in the "yyyy-mm-dd" format. This will be localized in `timezone` if given.
    stop_date : str, optional
        The end day up to which consider data in the "yyyy-mm-dd" format. This will be localized in `timezone` if given. This day will be INCLUDED.
    minAcc : int, optional
        The minimum accuracy for a point to be kept. If accuracy is larger than this the point will be discarded. **NOTE that the accuracy column must be called** ``acc``.
    sep : str, optional
        The delimiter of the fields in the files.
    file_schema : list of tuples
        The dict to rename the original columns to the mobilkit ones.
        **NOTE that the accuracy column must be called** ``acc``.

    Returns
    -------
    df : dask.dataframe
        A representation of the dataframe.
    '''
    files = [file for file in glob(pattern)]
    if header:
        sp_df = dd.read_csv(files, sep=sep)
    else:
        sp_df = dd.read_csv(files, names=file_schema.keys(), sep=sep)
    sp_df = sp_df.rename(columns=file_schema)
    selected_tz = pytz.timezone(timezone)
    sp_df = filterStartStopDates(sp_df, start_date, stop_date, selected_tz)
    sp_df = compute_datetime_col(sp_df, timezone)
    if minAcc is not None and minAcc > 0:
        print("Filtering on accuracy <= %d" % minAcc)
        sp_df = sp_df[sp_df[accColName] <= minAcc]
    return sp_df

def loaddata_takeapeek(dirpath, sep, ext):
    files = [file for file in glob(dirpath+"*"+ext)]
    for i,f in enumerate(files):
        print("filename of file #"+str(i)+": "+f)
        df = dd.read_csv(f, sep=sep)
        print("head of file #"+str(i)+":")
        print(df.head())

def compute_datetime_col(df, selected_tz):
    if dttColName in df.columns:
        print("Warning, '%s' column already present, not computing it..." % dttColName)
    else:
        df[dttColName] = dd.to_datetime(df[utcColName]*1e9, utc=False, unit="ns")\
                        .dt.tz_localize("UTC")\
                        .dt.tz_convert(selected_tz)\
                        .dt.tz_localize(None)
        df[dttColName] = df[dttColName].astype('datetime64[s]')

    return df

def localizeDatetimeNaive(date, tz, date_format="%Y-%m-%d"):
    dt = datetime.strptime(date, date_format)\
                        .astimezone(tz)\
                        .replace(tzinfo=None)
    return dt

def filterStartStopDates(df, start_date, stop_date, tz):
    if start_date is not None:
        start_UTC = localizeDatetimeNaive(start_date, tz).timestamp()
        print("Filtering on start UTC >= %d (%s)" % (start_UTC, start_date))
    else:
        start_UTC = -1
    if stop_date is not None:
        stop_UTC = localizeDatetimeNaive(stop_date, tz).timestamp()
        print("Filtering on stop UTC <= %d (%s)" % (stop_UTC, stop_date))
    else:
        stop_UTC = 99999999999999
    df = df[df[utcColName].between(start_UTC, stop_UTC)]
    return df

def load_from_skmob(df, uid="user", npartitions=10):
    '''
    Loads a dataframe imported with skmobility and returns a dask dataframe.

    Parameters
    ----------
    df : scikit-mobility.dataframe
        A dataframe as imported from scikit-mobility. May already contains the ``tile_ID`` and ``uid`` columns. If no ``uid`` column is found it will be initialized to the ``uid`` value.
    uid : str, optional
        The ``uid`` to be used, otherwise uses the present ones if the ``uid`` column is there.
    npartitions : int, optional
        The number of partition for the dataframe to be split into.

    Returns
    -------
    df_sp : dask.dataframe
        A `dask.dataframe` containing the input columns plus the accuracy ``acc`` (with dummy 1 value) and possibly the ``uid`` one if it was missing.


    '''
    import skmob
    # Create the acc column
    if accColName not in df.columns:
        df[accColName] = 1

    if uidColName not in df.columns:
        df[uidColName] = uid
    df_sp = dd.from_pandas(df, npartitions=npartitions)
    return df_sp

def dask_to_skmob(df, **kwargs):
    '''
    Ports a dataframe from dask to skmob. Given the structure of skmob it is done only to a `skmob.TrajDataFrame`.

    Parameters
    ----------
    df : dask.dataframe
        A dask dataframe with at least the ``uid``, ``lat`` and ``lng`` columns.
    **kwargs
        Will be passed to ``skmob.TrajDataFrame``.

    Returns
    -------
    df_sp : skmob.dataframe
        A skmob.TrajDataFrame containing the input columns.
    '''
    import skmob
    traj = skmob.TrajDataFrame(df.compute(), **kwargs)
    return traj

def persistDF(df, path, overwrite=True, header=True, index=False, out_format="csv"):
    '''
    Save a dask dataframo file.
    
    Parameters
    ----------
    df : dask.DataFrame
        The dataframe to save
    path : str
        The path where to save the dataframe.
    overwrite : bool
        Whether or not to force overwrite.
    header : bool
        Whether or not to put the header in the output file.
    index : bool
        Whether or not to put the index column in the output file.
    out_format : bool
        One of ``csv, parquet`` the format to use. If the df has arrays in it use ``parquet``.
    '''
    
    if not overwrite and os.path.exists(path):
        raise RuntimeError("Path %s already exist, force overwrite to continue." % path)

    if out_format == "csv":
        df.to_csv(path, header=header, index=index)
    elif out_format == "parquet":
        df.to_parquet(path, write_index=index, overwrite=overwrite)
    else:
        raise RuntimeError("Unknown format %s in persistDF" % out_format)

def reloadDF(path, header=True, in_format="csv"):
    '''
    Load a dask dataframe from file.
    
    Parameters
    ----------
    path : str
        The path where to read the dataframe from.
    header : bool
        Whether or not to read the header in the output file.
    in_format : bool
        One of ``csv, parquet`` the format used to persist the df.
    
    Returns
    -------
    df : dask.DataFrame
        The loaded dataframe.
    '''
    if in_format == "csv":
            df = dd.read_csv(path)
    elif in_format == "parquet":
            df = dd.read_parquet(path)
    else:
        raise RuntimeError("Unknown format %s in reloadDF" % out_format)
        
    return df

def fromunix2fulldate(x, timezone="America/Mexico_City"):
    '''
    Inherited from D4R.
    '''
    ct = pytz.timezone(timezone)
    ### errors 
    if (int(x)<86400) or (int(x)>1609518354):
        x = 86400
    dat = datetime.fromtimestamp(x, ct)
    tim = dat.strftime('%Y-%m-%d %H:%M:%S') ### date --> string
    tim2 = dat.strptime(tim, '%Y-%m-%d  %H:%M:%S') ### string --> date
    return tim2

def fromunix2date(x, timezone="America/Mexico_City"):
    '''
    Inherited from D4R.
    '''
    ct = pytz.timezone(timezone)
    ### errors 
    if (int(x)<86400) or (int(x)>1609518354):
        x = 86400
    dat = datetime.fromtimestamp(x, ct)
    tim = dat.strftime('%Y-%m-%d') ### date --> string
    tim2 = dat.strptime(tim, '%Y-%m-%d') ### string --> date
    return tim2

def fromunix2time(x, timezone="America/Mexico_City"):
    '''
    Inherited from D4R.
    '''
    ct = pytz.timezone(timezone)
    if (int(x)<86400) or (int(x)>1609518354):
        x = 86400
    dat = datetime.fromtimestamp(x, ct)
    tim = dat.strftime('%H:%M:%S')
    tim2 = dat.strptime(tim, '%H:%M:%S')
    return tim2

def crop_date(dff, startdt, enddt, timezone="America/Mexico_City"):
    ### input string start date 
    ct = pytz.timezone(timezone)
    st = int(datetime.strptime(startdt, "%Y/%m/%d").replace(tzinfo=ct).timestamp())
    et = int(datetime.strptime(enddt, "%Y/%m/%d").replace(tzinfo=ct).timestamp())
    df_res = dff[(dff[utcColName] <= et) & (dff[utcColName] >= st)]
    return df_res

def crop_time(dff, nighttime_start, nighttime_end, timezone): ### input string start date 
    ct = pytz.timezone(timezone)
    daytime1 =  datetime.strptime(nighttime_start, '%H:%M:%S')
    daytime2 =  datetime.strptime(nighttime_end, '%H:%M:%S')
    meta = dff[utcColName].head(1).apply(fromunix2time)
    dff["time_t"] = dff[utcColName].apply(lambda x: fromunix2time(x))
    dff_nighttime = dff[(dff["time_t"]<daytime1) | (dff["time_t"]>daytime2)]
    return dff_nighttime

def crop_spatial(dff, bbox):
    '''
    Filters `dff` with a `box=[minlon,minlat,maxlon,maxlat]`.
    '''
    minlon = bbox[0]
    minlat = bbox[1]
    maxlon = bbox[2]
    maxlat = bbox[3]
    df_res = dff[(dff[lonColName] <= maxlon) & (dff[lonColName] >= minlon) \
                 & (dff[latColName] <= maxlat) & (dff[latColName] >= minlat)]
    return df_res


def loadGeolifeData(path, acc_default=1, timezone="Asia/Shanghai"):
    '''
    Loads the `Geolife v1.3` trajectories with files ordered in the
    
        `Geolife\ Trajectories\ 1.3/Data/000/Trajectory/20090401202331.plt`
        
    structure with 6 useless rows at the beginning and the
    
        `lat,lng,0,altitude,days,date,time`
        
    format.
    
    Parameters
    ----------
    path : str
        The path to the root of the geolife data, usually called
        `data/Geolife\ Trajectories\ 1.3`.
    acc_default : float
        The default accuracy to give to each point to replicate the `mobilkit`
        format.
    timezone : str
        The code of the timezone the data has been recorded in.
    
    Returns
    -------
    df : pd.DataFrame
        The dataframe containing the
            `uid,UTC,datetime,acc,lat,lng` columns.
    '''
    
    folders_pattern = os.path.join(path, "Data/*")
    user_dirs = [d for d in sorted(glob(folders_pattern)) if os.path.isdir(d)]
    df = pd.DataFrame()
    n_users = len(user_dirs)
    sys.stdout.write("Found %d users...\n" % n_users)
    sys.stdout.flush()
    from multiprocessing import Pool, cpu_count
    world = cpu_count()
    with Pool(world) as p:
        dfs = p.map(_loaderGeoLifeBatch, [(user_dirs, acc_default, timezone, i, world) for i in range(world)])
    df = pd.concat(dfs, ignore_index=True, sort=True)
    sys.stdout.write("\nDone!")
    sys.stdout.flush()
    
    return df

def _loaderGeoLifeBatch(args):
    user_dirs, acc_default, timezone, rank, world = args
    df = pd.DataFrame()
    for user_dir in user_dirs[rank::world]:
        tmp_user = int(os.path.basename(user_dir))
        traj_pattern = os.path.join(user_dir, "Trajectory/*.plt")
        for traj in sorted(glob(traj_pattern)):
            if not os.path.isfile(traj):
                continue
            tmp_rows = pd.read_csv(traj,sep=",", skiprows=6, header=None)
            tmp_rows.columns = [latColName,lonColName,"OS","junk_1","junk_2","day","time"]
            tmp_rows["datetime_str"] = tmp_rows.apply(lambda r: r["day"] + " " + r["time"], axis=1)
            tmp_rows[uidColName] = [tmp_user]*tmp_rows.shape[0]
            tmp_rows[accColName] = [acc_default]*tmp_rows.shape[0]
            tmp_rows[dttColName] = pd.to_datetime(tmp_rows["datetime_str"], utc=True,
                                                  format="%Y-%m-%d %H:%M:%S")\
                                        .dt.tz_convert(timezone)

            tmp_rows[utcColName] = (tmp_rows[dttColName].dt.tz_convert("GMT")
                                    - pd.Timestamp("1970-01-01", tz="GMT"))\
                                            // pd.Timedelta('1s')

            tmp_rows = tmp_rows[[uidColName,utcColName,dttColName,accColName,latColName,lonColName]]
            df = pd.concat((df, tmp_rows), ignore_index=True, sort=True)
    sys.stdout.write("Done process %03d / %03d\n" % (rank+1, world))
    sys.stdout.flush()
    return df


def syntheticGeoLifeDay(df_geolife, selected_day):
    # Many users, one day aggregation
    # Each user/{weekday/weekend} is a different user
    df_users = df_geolife.copy()

    df_users["day"] = df_users[dttColName].dt.floor("1d")

    seen_couples = df_users.groupby([uidColName,"day"])[[]].agg("count").reset_index()
    target_col = uidColName + "_new"
    seen_couples[target_col] = np.arange(0, seen_couples.shape[0])
    df_users = pd.merge(df_users, seen_couples, on=[uidColName,"day"], how="inner")
    df_users[uidColName] = df_users[target_col].values
    del df_users[target_col], df_users["day"]

    df_users[dttColName] = selected_day + pd.to_timedelta(df_users[dttColName].dt.hour*3600
                                                            + df_users[dttColName].dt.minute*60
                                                            + df_users[dttColName].dt.second, unit="s")

    df_users[utcColName] = (df_users[dttColName].dt.tz_convert("GMT")
                                        - pd.Timestamp("1970-01-01", tz="GMT"))\
                                                // pd.Timedelta('1s')
    
    return df_users

def syntheticGeoLifeWeek(df_geolife, selected_week):
    while selected_week.weekday() > 0:
        selected_week -= timedelta(days=1)
        if selected_week.weekday() == 0:
            print("Anticipated the date to Monday: ", selected_week)
            
    # Less users, one week aggregation
    # Each user/week is a different user, all projected to a given week
    df_week = df_geolife.copy()

    df_week["week"] = df_week[dttColName].dt.floor("1d") - pd.to_timedelta(df_week[dttColName].dt.weekday, unit="d")

    seen_couples = df_week.groupby([uidColName,"week"])[[]].agg("count").reset_index()
    target_col = uidColName + "_new"
    seen_couples[target_col] = np.arange(0, seen_couples.shape[0])
    df_week = pd.merge(df_week, seen_couples, on=[uidColName,"week"], how="inner")
    df_week[uidColName] = df_week[target_col].values
    del df_week[target_col], df_week["week"]

    df_week[dttColName] = selected_week + pd.to_timedelta(df_week[dttColName].dt.weekday*(24*3600)
                                                            + df_week[dttColName].dt.hour*3600
                                                            + df_week[dttColName].dt.minute*60
                                                            + df_week[dttColName].dt.second, unit="s")

    df_week[utcColName] = (df_week[dttColName].dt.tz_convert("GMT")
                                        - pd.Timestamp("1970-01-01", tz="GMT"))\
                                                // pd.Timedelta('1s')
    
    return df_week



# Copyright (C) MindEarth <enrico.ubaldi@mindearth.org> @ Mindearth 2020-2021
# 
# This file is part of mobilkit.
#
# mobilkit is distributed under the MIT license.

'''Tools and functions to spatially analyze the data.


.. note::
    When determining the home location of a user, please consider that some data providers, like *Cuebiq*, obfuscate/obscure/alter the coordinates of the points falling near the user's home location in order to preserve privacy.

    This means that you cannot locate the precise home of a user with a spatial resolution higher than the one used to obfuscate these data. If you are interested in the census area (or geohash) of the user's home alone and you are using a spatial tessellation with a spatial resolution wider than or equal to the one used to obfuscate the data, then this is of no concern.

    However, tasks such as stop-detection or POI visit rate computation may be affected by the noise added to data in the user's home location area. Please check if your data has such noise added and choose the spatial tessellation according to your use case.

'''
from dask import dataframe as dd
from dask import array as da
import dask.bag as db

from mobilkit.dask_schemas import nunique

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib
from copy import copy, deepcopy
import geopandas as gpd
from scipy.spatial import Voronoi
import shapely
from sklearn import cluster
from sklearn.metrics.pairwise import haversine_distances
from sklearn.cluster import DBSCAN

from mobilkit.dask_schemas import (
    accColName,
    lonColName,
    latColName,
    uidColName,
    utcColName,
    dttColName,
    zidColName,
    stpColName,
    ldtColName,
    medLatColName,
    medLonColName,
    locColName,
    radLatColName,
    radLonColName,
    durColName,
)
from mobilkit.tools import computeGDFbounds

from mpl_toolkits.axes_grid1 import make_axes_locatable
from itertools import chain
from shapely.geometry import Polygon

from sklearn.neighbors import KDTree
from scipy.spatial.distance import euclidean


def tessellate(df, tesselation_shp,
               filterAreas=False,
               partitions_number=None,
               latCol=latColName,
               lonCol=lonColName,
               ):
    '''Function to assign to each point a given area index.

    Parameters
    ----------
    df : dask.DataFrame
        A dataframe as returned from :attr:`mobilkit.loader.load_raw_files` or imported from ``scikit-mobility`` using :attr:`mobilkit.loader.load_from_skmob`.
    tesselation_shp : str
        The path (relative or absolute) to the shapefile containing the tesselation of space. If the shapefile does not contain a `tile_ID` field it will be initialized here and included in the returned geodataframe.
    filterAreas : bool
        If tesselation is specified, keeps only the points within the specified shapofile.
    partitions_number : int, optional
        The batch size of the geopandas sjoin function to be applied. Leave it as is unless you know what you're doing.
    latCol, lonCol : str, optional
        The names of the columns containing the latitude and longitude coordinates.

    Returns
    -------
    df_tile : dask.dataframe
        The initial dataframe with the additional ``tile_ID`` column telling the int id of the area the point is belonging to (-1 if the point is outside of the shapefile bounds).
    tessellation_gdf : geopandas.GeoDataFrame
        The geo-dataframe with the possibly missing `tile_ID` column added.
    '''

    zones_gdf = gpd.read_file(tesselation_shp)
    if zidColName not in zones_gdf.columns:
        zones_gdf[zidColName] = np.arange(zones_gdf.shape[0], dtype=int)
    
    # This is the pipeline, it basically combines all the steps and write out the info
    def localAssign(_df):
        tmp_gdf = gpd.GeoDataFrame(_df[[latCol]],
                               geometry=gpd.points_from_xy(_df[lonCol],
                                                            _df[latCol]))
        tmp_gdf.crs = zones_gdf.crs
        df_out = gpd.sjoin(tmp_gdf[["geometry"]],
                            zones_gdf[[zidColName, "geometry"]], how="left")
        df_out = df_out[~df_out.index.duplicated()]
        df_out[zidColName].fillna(-1, inplace=True)
        _df = _df.join(df_out[[zidColName]].astype(int))
        return _df

    if filterAreas:
        # Leave at once the pings outside of bounds
        bounds = computeGDFbounds(zones_gdf)
        minx, maxx, miny, maxy = (
            bounds['minx'],
            bounds['maxx'],
            bounds['miny'],
            bounds['maxy'],
        )
        query = '{0} >= {2} & {0} <= {3} & {1} >= {4} & {1} <= {5}'.format(lonCol,
                                                                           latCol,
                                                                           minx, maxx,
                                                                           miny, maxy)
        df = df.query(query).copy()

    if partitions_number is not None:
        df = df.repartition(npartitions=partitions_number)
    out_df = df.map_partitions(localAssign)
    
    if filterAreas:
        out_df = out_df[out_df[zidColName] >= 0]

    return out_df, zones_gdf


def assignAreasDF(df, zones_gdf):
    '''Returns the geo-dataframe with an additional column the `ZONE_IDX` column.
    Non overlapping areas are guaranteed to be found there with a negative -1 value;
    The order of the original index and columns is preserved.
    '''
    # Assign stop to areas
    tmp_gdf = gpd.GeoDataFrame(df[[latColName]],
                               geometry=gpd.points_from_xy(df[lonColName],
                                                            df[latColName]))
    tmp_gdf.crs = zones_gdf.crs
    df_out = gpd.sjoin(tmp_gdf[["geometry"]],
                        zones_gdf[[zidColName, "geometry"]], how="left")
    df_out = df_out[~df_out.index.duplicated()]
    df_out[zidColName].fillna(-1, inplace=True)
    df[zidColName] = df_out[zidColName].astype(int)
    return df


def plotHomeWorkPoints(uid, df_hw, gdf, ax=None, kwargs_bounds=None, kwargs_points=None):
    '''Plots the points in home and work hours for an user on the map.

    Parameters
    ----------
    uid : (str or int, depending on the uid type)
        The id of the user to plot.
    df_hw : dask.dataframe
        A dataframe as returned by :attr:`mobilkit.stats.userHomeWork` with at least the `uid`, `tile_ID` and `isHome` and `isWork` columns.
    gdf : geopandas.GeoDataFrame
        A geo-dataframe as returned by :attr:`mobilkit.spatial.tessellate`.
    ax : pyplot.axes, optional
        The axes where to plot. If ``None`` (default) creates a new figure.
    kwargs_bounds : dict, optional
        Will be passed to the geopandas plot function plotting the boundaries.
    kwargs_bounds : dict, optional
        Will be passed to the geopandas plot function plotting the boundaries.

    Returns
    -------
    ax : pyplot.axes, optional
        The axes of the figure.
    '''

    if ax is None:
        fig, ax = plt.subplots(1,1,figsize=(12,12))
    ax.set_aspect("equal")

    kw_bd = dict(color="None", edgecolor="k", alpha=.8, lw=2)
    if kwargs_bounds is not None:
        for k, v in kwargs_bounds.items():
            kw_bd[k] = v

    kw_pt = dict(legend=True, cmap="Set1")
    if kwargs_points is not None:
        for k, v in kwargs_points.items():
            kw_pt[k] = v

    ax = gdf.plot(ax=ax, **kw_bd)

    df_usr = df_hw[df_hw[uidColName] == uid].compute()
    points_u = gpd.GeoDataFrame(df_usr,
            geometry=gpd.points_from_xy(df_usr[lonColName],df_usr[latColName]))
    points_u["kind"] = None
    points_u.loc[points_u["isHome"]==1,"kind"] = "Home"
    points_u.loc[points_u["isWork"]==1,"kind"] = "Work"
    points_plt = points_u[~points_u["kind"].isna()].copy()

    ax = points_plt.plot("kind", ax=ax, **kw_pt)

    ax.set_xlim(points_u[lonColName].min()-.01, points_u[lonColName].max()+.01)
    ax.set_ylim(points_u[latColName].min()-.01, points_u[latColName].max()+.01)

    return ax

def plotHomeWorkUserCount(df_hw_locs, gdf, what="home", ax=None, kwargs_map=None):
    '''Plots a colormap of the number of people living (or working) in each area.

    Parameters
    ----------
    df_hw_locs : pandas.dataframe
        A dataframe as returned by :attr:`mobilkit.stats.userHomeWorkLocation` with at least the `uid`, `home_tile_ID` `work_tile_ID` columns and passed to pandas.
    gdf : geopandas.GeoDataFrame
        A geo-dataframe as returned by :attr:`mobilkit.spatial.tessellate`.
    what :  str
        The ``home`` or ``work`` string, telling whether to plot the number of people living or working in an area.
    ax : pyplot.axes, optional
        The axes where to plot. If ``None`` (default) creates a new figure.
    kwargs_map : dict, optional
        Will be passed to the geopandas plot function plotting the boundaries and colormap.

    Returns
    -------
    ax : pyplot.axes, optional
        The axes of the figure.
    gdf : geopandas.GeoDataFrame
        The original geo dataframe with an additional column (``n_users_home`` if counting home or ``n_users_work`` if counting work). If the column is already in the df it will be overwritten.
    df : pandas.DataFrame
        The ``tile_ID`` -> count of users mapping.
    '''

    assert what in ["home", "work"]

    # Count the number of people working or living in an area

    columnToUse = "home_tile_ID" if what == "home" else "work_tile_ID"
    targetColumn = "n_users_home" if what == "home" else "n_users_work"
    plotLabel = "Number of residents" if what == "home" else "Number of workers"

    locs_hh = df_hw_locs.dropna(subset=[columnToUse])\
                    .groupby(columnToUse)\
                    .agg({uidColName: 'nunique'})

    locs_hh = locs_hh.reset_index().rename(columns={uidColName: targetColumn,
                                                    columnToUse: zidColName})
    locs_hh[zidColName] = locs_hh[zidColName].astype(int)

    if targetColumn in gdf.columns:
        del gdf[targetColumn]

    gdf = pd.merge(gdf, locs_hh, on=zidColName, how="left")
    gdf[targetColumn] = gdf[targetColumn].fillna(0)

    if ax is None:
        fig, ax = plt.subplots(1,1,figsize=(12,12))
    ax.set_aspect("equal")

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.1)

    kw_plot = dict(edgecolor="none", lw=2, cax=cax, ax=ax, legend=True)
    if kwargs_map is not None:
        for k, v in kwargs_map.items():
            kw_plot[k] = v

    ax = gdf.plot(targetColumn, **kw_plot)

    cax.set_ylabel(plotLabel, size=22)
    cax.yaxis.set_tick_params(labelsize=18)

    ax.axis("off")

    return ax, gdf, locs_hh


def plotActivityCount(df_act, gdf, what="pings", ax=None, kwargs_map=None):
    '''Plots a colormap of the number of pings (or unique users) observed in a given area in a given period.

    Parameters
    ----------
    df_act : pandas.dataframe
        A dataframe as returned by :attr:`mobilkit.spatial.areaStats` with at least the `tile_ID` and `pings` and/or `users` columns and passed to pandas.
    gdf : geopandas.GeoDataFrame
        A geo-dataframe as returned by :attr:`mobilkit.spatial.tessellate`.
    what :  str
        The ``pings`` or ``users`` string, telling whether to plot the number of pings recorded in an area or the number of unique users seen there.
    ax : pyplot.axes, optional
        The axes where to plot. If ``None`` (default) creates a new figure.
    kwargs_map : dict, optional
        Will be passed to the geopandas plot function plotting the boundaries and colormap.

    Returns
    -------
    ax : pyplot.axes, optional
        The axes of the figure.
    '''


    if what == "pings":
        plotLabel = "Pings count"
    elif what == "users":
        plotLabel = "Unique users count"
    else:
        raise RuntimeError("Unknown ``what`` '%s' in ``plotActivityCount``" % what)

    df_act = df_act[[zidColName, what]]

    if what in gdf.columns:
        del gdf[what]

    gdf = pd.merge(gdf, df_act, on="tile_ID", how="left")
    gdf[what] = gdf[what].fillna(0)

    if ax is None:
        fig, ax = plt.subplots(1,1,figsize=(12,12))
    ax.set_aspect("equal")

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.1)

    kw_plot = dict(edgecolor="none", lw=2, cax=cax, ax=ax, legend=True)
    if kwargs_map is not None:
        for k, v in kwargs_map.items():
            kw_plot[k] = v

    ax = gdf.plot(what, **kw_plot)

    cax.set_ylabel(plotLabel, size=22)
    cax.yaxis.set_tick_params(labelsize=18)

    ax.axis("off")

    return ax


def selectAreasFromBounds(gdf, relation="within", min_lon=-99.15913, max_lon=-99.10032, min_lat=19.41353, max_lat=19.46100,):
    '''Function to select areas from a geodataframe given the bounds of a selected region.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        A geodataframe with at least the ``tile_ID`` and ``geometry`` columns as returned by :attr:`mobilkit.spatial.tessellate`.
    relation : str, optional
        The relation between the bounds and the areas. "within" or "intersects"
    min/max_lon/lat : float, optional
        The minimum and maximum latitude and longitude of the box.

    Returns
    -------
    areas_ids : set
        The set of the areas **within** or **intersecting** the given bounds
    '''

    assert relation in ["within", "intersects"]

    selected_polygon = Polygon([
            [min_lon, min_lat],
            [min_lon, max_lat],
            [max_lon, max_lat],
            [max_lon, min_lat],
            [min_lon, min_lat]
        ])

    if relation == "within":
        idxs = gdf.within(selected_polygon)
    elif relation == "intersects":
        idxs = gdf.intersects(selected_polygon)
    return set([int(i) for i in gdf[idxs][zidColName].values])


def replaceAreaID(df, mapping):
    '''Function that replaces all the ``tile_ID`` with a new id given in the mapping.

    Parameters
    ----------
    df : dask.DataFrame
        A dataframe with at least the ``tile_ID`` column.
    mapping : dict
        A mapping between the original ``tile_ID`` and the new desired one. MUST CONTAIN ALL THE ``tile_ID`` s present in df.

    Returns
    -------
    df_out : dask.DataFrame
        A copy of the original dataframe with the ``tile_ID`` replaced.

    '''
    df = df.map_partitions(lambda p: p.replace({zidColName: mapping}))
    return df


def meanshift(df, bw=0.01, maxpoints=100, **kwargs):
    '''
    Given the points of a user finds the home location with MeanShift
    clustering.
    
    Parameters
    ----------
    df : pandas.DataFrame
        With at least `latcol,loncol`.
    bw : float
        Bandwidth to be used in MeanShift.
    maxpoints : int
        The maximum number of points to be used in meanshift.
        If more, a fraction of the df to have `maxpoints` will be sampled.
    kwargs
        Will be passed to `sklearn.cluster.MeanShift` constructor.
        
    Returns
    -------
    clust_center : tuple
        The center of the cluster found in the `(longitude,latitude)` format.
     
    Note
    ----
    When determining the home location of a user, please consider that some data providers, like *Cuebiq*, obfuscate/obscure/alter the coordinates of the points falling near the user's home location in order to preserve privacy.

    This means that you cannot locate the precise home of a user with a spatial resolution higher than the one used to obfuscate these data. If you are interested in the census area (or geohash) of the user's home alone and you are using a spatial tessellation with a spatial resolution wider than or equal to the one used to obfuscate the data, then this is of no concern.

    However, tasks such as stop-detection or POI visit rate computation may be affected by the noise added to data in the user's home location area. Please check if your data has such noise added and choose the spatial tessellation according to your use case.
    '''
    if len(df)>maxpoints:
        df = df.sample(frac=maxpoints/df.shape[0])
    ms = cluster.MeanShift(bandwidth=bw, **kwargs)
    ms.fit(df[[lonColName,latColName]])
    labels = ms.labels_
    counts = np.bincount(labels)
    most = np.argmax(counts)
    cluster_centers = ms.cluster_centers_
    return cluster_centers[most]

def userHomeWorkDistance(r):
    '''
    Computes the distance between `lat/lng_home` and `lat/lng_work`
    for the row of a user.

    Returns
    -------
    dist : float
        None if one of the coords is invalid, the distance in km otherwise.
    '''
    if pd.isna(r[lonColName+'_home']) or pd.isna(r[lonColName+'_work']):
        return None
    else:
        home_place = np.array([r[latColName+'_home'], r[lonColName+'_home']]).reshape((1,2))
        work_place = np.array([r[latColName+'_work'], r[lonColName+'_work']]).reshape((1,2))
        return haversine_pairwise(home_place, work_place)[0,0]

def haversine_pairwise(X,
                       Y=None,
                       isRadians=False):
    '''
    Parameters
    ----------
    X, Y : np.array
        a Nx * 2 and Ny*2 arrays of (lat,lon) coordinates.
        If `Y` is `None` it will be assigned to `X` (computes the matrix of distances of `X` items).
    isRadians : bool, optional
        Whether the supplied coordinates are already in radians. If not they will be automatically converted.
    
    Returns
    -------
    distances : np.array
        a Nx*Ny matrix of distances in kilometers
    '''
    if not isRadians:
        X = np.radians(X)
    if Y is None:
        Y = X
    else:
        if not isRadians:
            Y = np.radians(Y)
        
    distances = haversine_distances(X, Y)
    distances *= 6372.795 # multiply by Earth radius in km to get kilometers
    
    return distances

# Stops and stays
def _stopRange(row, interval='1h'):
    '''
    Helper function that takes a row with :attr:`mobilkit.dask_schemas.dttColName`
    and :attr:`mobilkit.dask_schemas.ldtColName` as input and returns the times
    between the initial and final time. If the end time crosses the interval time bin
    the result is guaranteed to have two (or more) entries (e.g. with a 1 hour interval,
    an initial time at 08:34 and a final time at 09:01 we find two steps at 08:34 and 09:34).
    '''
    t_i = row[dttColName]
    t_f = row[ldtColName].ceil(interval) - np.timedelta64(int(1e6))
    return pd.date_range(t_i, t_f, freq=interval)


def _expand_stops_partition(df, freq='1h', explode_stop=False):
    '''
    Auxiliary function that takes as input a df with at least the
    :attr:`mobilkit.dask_schemas.dttColName` and :attr:`mobilkit.dask_schemas.ldtColName` 
    columns and computes all the times in between spaced by `freq`. The result is
    written in the :attr:`mobilkit.dask_schemas.stpColName` and possibly exploded if
    `explode_stop=True`.
    '''
    if df.shape[0] == 0:
        df[stpColName] = []
    else:
        df[stpColName] = df.apply(_stopRange, axis=1, interval=freq)
    if explode_stop:
        df = df.explode(stpColName)
    del df[ldtColName]
    return df


def expandStops(df, freq:str='1h', explode_stop:bool=True):
    '''
    Given a dataframe containing the single stops of one or more users (as returned by
    :attr:`mobilkit.spatial.findStops`) it explodes them to be repeated once every `freq`
    time bin they traverse (if `explode_stop`) or it lists them in the
    :attr:`mobilkit.dask_schemas.stpColName` column. 
    
    Parameters
    ----------
    df : dask.datframe
        A dataframe containing the stops of one or more users (as returned by
        :attr:`mobilkit.spatial.findStops`).
    freq : str, optional
        The time bin to use to replicate/explode the stop. Currently all the valid
        `freq` arguments to `pandas.date_range` are accepted.
    explode_stop : bool, optional
        If `True` the :attr:`mobilkit.dask_schemas.stpColName` column will contain
        the exploded list of time bins touched by the stop, else the list itself.

    Returns
    -------
    exploded_stops_df : dask.datframe
        The initial dataframe with the additional column :attr:`mobilkit.dask_schemas.stpColName`
        containing the time bins (or their list, depending on `explode_stop`) touched
        by the stop.
    '''
    return_meta = dict(**df.dtypes)
    if explode_stop:
        return_meta[stpColName] = return_meta[dttColName]
    else:
        return_meta[stpColName] = object # If explode stops is as dttcolname
    exploded_stops_df = df.map_partitions(_expand_stops_partition,
                                            freq=freq,
                                            explode_stop=explode_stop)
    return exploded_stops_df


def _find_stops(df, custom_stay_locations_kwds=None):
    '''
    Auxiliary function to be applied to the per-user slice of a grouped
    df containing the ping.
    If not told otherwise, the stay location keywords passed to
    `skmob.preprocessing.detection` are:
    - `minutes_for_a_stop=5.0`
    - `spatial_radius_km=0.2`
    - `no_data_for_minutes=60*12`
    - `leaving_time=True`
    Theleaving datetime of the stop is saved into the :attr:`mobilkit.dask_schemas.ldtColName`
    column.
    '''
    stay_locations_kwds = dict(minutes_for_a_stop=5.0,
                                       spatial_radius_km=0.2,
                                       no_data_for_minutes=60*12,
                                       leaving_time=True)
    if custom_stay_locations_kwds:
        stay_locations_kwds.update(custom_stay_locations_kwds)

    from skmob import TrajDataFrame
    from skmob.preprocessing import detection
    tdf = TrajDataFrame(df[[latColName, lonColName, uidColName, dttColName]],
                              latitude=latColName,
                              longitude=lonColName,
                              datetime=dttColName,
                              trajectory_id=None,
                              user_id=uidColName)
    
    stdf = detection.stay_locations(tdf, **stay_locations_kwds)
    stdf = pd.DataFrame(stdf)    
    return stdf


def findStops(df,
            tesselation_shp=None,
            stay_locations_kwds=None,
            filterAreas=True,
            ):
    '''
    Computes the stops of a group of users using the `scikit-mobility` tools.
    Note that the `mobilkit[complete]` or `[skmob]` version should be installed to use this tool.
    
    Parameters
    ----------
    df : dask.dataframe
        A dataframe containing the raw pings with at least the latitude, longitude and datetime columns.
    tessellation_shp : str, optional
        The path to be used to tessellate the stops (if `None`, no tessellation will be performed).
    stay_locations_kwds : dict, optional
        The custom keywords to be passed to :attr:`mobilkit.spatial._find_stops`. If not told otherwise,
        the stay location keywords passed to `skmob.preprocessing.detection` are:
            - `minutes_for_a_stop=5.0`
            - `spatial_radius_km=0.2`
            - `no_data_for_minutes=60*12`
            - `leaving_time=True`
    filterAreas : bool, optional
        Whether or not to filter out stops found outside of the tessellation when tessellating.
    
    Returns
    -------
    stops_df : dask.dataframe
        The dataframe with the latitude and longitude of each stop together with:
        - its starting time (in the :attr:`mobilkit.dask_schemas.dttColName`)
        - its ending time (saved into the :attr:`mobilkit.dask_schemas.ldtColName` column).
        - the duration of the stop in seconds in the :attr:`mobilkit.dask_schemas.durColName` column.
        - if a tessellation file is specified, an additional :attr:`mobilkit.dask_schemas.zidColName`
          is telling in which grid cell the stop is falling.
    '''
    # return_meta = dict(**df_pings_filtered.dtypes)
    return_meta = dict(**df.dtypes)
    return_meta = {k: return_meta[k] for k in [latColName, lonColName,
                                               uidColName, dttColName]}
    return_meta[ldtColName] = return_meta[dttColName]
    stops_df = df.groupby(uidColName)\
                            .apply(_find_stops,
                                   meta=return_meta,
                                   custom_stay_locations_kwds=stay_locations_kwds,
                                   ).map_partitions(lambda d: d.reset_index(drop=True))\
                            .repartition(npartitions=200)
    stops_df[durColName] = (stops_df[ldtColName]
                             - stops_df[dttColName]).dt.total_seconds()
    if tesselation_shp:
        stops_df, _ = tessellate(stops_df, tesselation_shp, filterAreas=filterAreas)
    return stops_df

def computeUsersLocations(stops_df,
                          method='dbscan',
                          link_dist=150,
                          min_stops_count=2,
                          return_locations=True,
                         ):
    '''
    TODO DOC
    '''
    out_meta = stops_df.dtypes
    out_meta = out_meta.to_dict()
    out_meta[locColName] = int
    out_meta[medLonColName] = float
    out_meta[medLatColName] = float
    
    if method == 'dbscan':
        stops_locs_df = stops_df.groupby(uidColName)\
                            .apply(_find_user_locations_DBSCAN,
                                   db_eps=link_dist,
                                   db_min_count=min_stops_count,
                                   meta=out_meta,
                            ).reset_index()
    elif method == 'infostop':
        try:
            import infostop
        except ModuleNotFoundError:
            raise RuntimeError('Module infostop not installed, please intall it with `pip install mobilkit[locations]`')
        # raise RuntimeError('Location detection method %s TODO!' % method)
        stops_locs_df = stops_df.groupby(uidColName)\
                            .apply(_find_user_locations_INFOSTOP,
                                   label_singleton=min_stops_count==1,
                                   meta=out_meta,
                            ).reset_index()
    else:
        raise RuntimeError('Location detection method %s not implemented!' % method)
    
    if return_locations:
        locs_users_df = stops_locs_df.groupby([uidColName,locColName])\
                                .agg({
                                    medLatColName: 'first',
                                    medLonColName: 'first',
                                    durColName: 'sum',
                                })
        return stops_locs_df, locs_users_df
    else:
        return stops_locs_df

# Locations dbscan
def _find_user_locations_DBSCAN(df_stops,
                                db_eps=100.,
                                db_min_count=1,
                                add_db_kws={},
                                latColName=latColName,
                                lonColName=lonColName,
                                returnLocs=False,
                               ):
    '''
    Computes the locations of a single user whose stops are in `df_stops`
    using the DBscan algorithm.
    
    Parameters
    ----------
    df : dask.DataFrame or pd.DataFrame
        The dataframe with at least the `latColName` and `lonColName`.
    db_eps : float, optional
        The distance, in meters, to link two stops in a cluster.
    db_min_count : int, optional
        The minimum count of neighbors to become core in the DBscan algorithm.
    add_db_kws : dict, optional
        Default keyworkds, besides `eps` and `min_count` are:
        - metric="haversine",
        - algorithm="ball_tree"
    latColName, lonColName : str, optional
        The columns of `df` containing the latitude and longitude.
    returnLocs : bool, optional
        If True returns the stops annotated with the
        - :attr:`mobilkit.dask_schemas.medLatColName`
        - :attr:`mobilkit.dask_schemas.medLonColName`
        - :attr:`mobilkit.dask_schemas.locColName`
        and a df with the unique locations of the user, otherwise
        only the annotated stops dataframe.
        
    Returns
    -------
    stops_df, user_locs : pd.Series
        A series with columns :attr:`mobilkit.dask_schemas.medLatColName` and
        :attr:`mobilkit.dask_schemas.medLonColName` containing the latitude and
        longitude of the medoid.
    '''
    db_kws = dict(
        metric="haversine",
        algorithm="ball_tree"
    )
    db_kws.update(add_db_kws)
    
    if df_stops.shape[0] > 0:
        df_stops[radLonColName] = np.radians(df_stops[lonColName])
        df_stops[radLatColName] = np.radians(df_stops[latColName])

        EARTH_RADIUS = 6372.795 * 1000
        eps = db_eps / EARTH_RADIUS
        
        DB_model = DBSCAN(eps=eps,
                          min_samples=db_min_count,
                          **db_kws)
        labels = DB_model.fit_predict(df_stops[[radLatColName, radLonColName]].values)
        del df_stops[radLatColName], df_stops[radLonColName]
        df_stops[locColName] = labels
        
        # Computing Clusters
        clust_med = df_stops.groupby(locColName).apply(points_to_medoid)
        df_stops[medLonColName] = df_stops.apply(lambda c:
                                                 clust_med.loc[c[locColName],medLonColName]
                                                     if c[locColName] >= 0
                                                         else c[lonColName],
                                                 axis=1)
        df_stops[medLatColName] = df_stops.apply(lambda c:
                                                 clust_med.loc[c[locColName],medLatColName]
                                                     if c[locColName] >= 0
                                                         else c[latColName],
                                                 axis=1)
    else:
        # No points
        clust_med = None
        labels = []
        lat_medoid = []
        lon_medoid = []
        df_stops[locColName] = labels
        df_stops[medLonColName] = lon_medoid
        df_stops[medLatColName] = lat_medoid
    if returnLocs:
        return df_stops, clust_med
    else:
        return df_stops


def _find_user_locations_INFOSTOP(df_stops,
                                r2=100.,
                                label_singleton=False,
                                add_infostop_kws={},
                                latColName=latColName,
                                lonColName=lonColName,
                                returnLocs=False,
                               ):
    '''
    Computes the locations of a single user whose stops are in `df_stops`
    using the infostop algorithm on the direct user trace.
    
    # TODO merge this and the dbscan function in a unique function as only the labelling changes
    Parameters
    ----------
    df : dask.DataFrame or pd.DataFrame
        The dataframe with at least the `latColName` and `lonColName`.
    db_eps : float, optional
        The distance, in meters, to link two stops in a cluster.
    db_min_count : int, optional
        The minimum count of neighbors to become core in the DBscan algorithm.
    add_db_kws : dict, optional
        Default keyworkds, besides `eps` and `min_count` are:
        - metric="haversine",
        - algorithm="ball_tree"
    latColName, lonColName : str, optional
        The columns of `df` containing the latitude and longitude.
    returnLocs : bool, optional
        If True returns the stops annotated with the
        - :attr:`mobilkit.dask_schemas.medLatColName`
        - :attr:`mobilkit.dask_schemas.medLonColName`
        - :attr:`mobilkit.dask_schemas.locColName`
        and a df with the unique locations of the user, otherwise
        only the annotated stops dataframe.
        
    Returns
    -------
    stops_df, user_locs : pd.Series
        A series with columns :attr:`mobilkit.dask_schemas.medLatColName` and
        :attr:`mobilkit.dask_schemas.medLonColName` containing the latitude and
        longitude of the medoid.
    '''
    from infostop import SpatialInfomap

    is_kws = dict()
    is_kws.update(add_infostop_kws)
    
    if df_stops.shape[0] > 0:
        IS_model = SpatialInfomap(r2=r2, label_singleton=label_singleton)
        labels = IS_model.fit_predict(df_stops[[latColName, lonColName]].values)
        df_stops[locColName] = labels
        
        # Computing Clusters
        clust_med = df_stops.groupby(locColName).apply(points_to_medoid)
        df_stops[medLonColName] = df_stops.apply(lambda c:
                                                 clust_med.loc[c[locColName],medLonColName]
                                                     if c[locColName] >= 0
                                                         else c[lonColName],
                                                 axis=1)
        df_stops[medLatColName] = df_stops.apply(lambda c:
                                                 clust_med.loc[c[locColName],medLatColName]
                                                     if c[locColName] >= 0
                                                         else c[latColName],
                                                 axis=1)
    else:
        # No points
        clust_med = None
        labels = []
        lat_medoid = []
        lon_medoid = []
        df_stops[locColName] = labels
        df_stops[medLonColName] = lon_medoid
        df_stops[medLatColName] = lat_medoid
    if returnLocs:
        return df_stops, clust_med
    else:
        return df_stops 

def points_to_medoid(df,
                     latColName=latColName,
                     lonColName=lonColName,
                    ):
    '''
    Returns the pd.Series containing the latitude and longitude of the medoid for
    the current df.
    
    Parameters
    ----------
    df : dask.DataFrame or pd.DataFrame
        The dataframe with at least the `latColName` and `lonColName`.
    latColName, lonColName : str, optional
        The columns of `df` containing the latitude and longitude.
        
    Returns
    -------
    medoid : pd.Series
        A series with columns :attr:`mobilkit.dask_schemas.medLatColName` and
        :attr:`mobilkit.dask_schemas.medLonColName` containing the latitude and
        longitude of the medoid.
    '''
    tmp_latlon = df[[latColName, lonColName]].values
    dist_matrix = haversine_pairwise(tmp_latlon)
    medoid_index = compute_medoid_index(dist_matrix)
    return pd.Series({
        medLatColName: df[latColName].iloc[medoid_index],
        medLonColName: df[lonColName].iloc[medoid_index],
    })

def compute_medoid_index(distM):
    '''
    Returns the row index of the necessarily symmetric matrix that
    minimizes the sum of distances to all the other points.

    Parameters
    ----------
    distM : np.array
        The distances matrix.

    Returns
    -------
    index : int
        The row (column) index of the medoid.

    '''
    return np.argmin(np.sum(distM, axis=0))

# Distances and radius of gyration
def compute_ROG(df, which='both', df_hw_locs=None):
    meta_out_rog = {'n_pings': np.int64}
    if which in ['home','both']:
        meta_out_rog.update({
        'rog_home': np.float64,
        'com_home_'+latColName: np.float64,
        'com_home_'+lonColName: np.float64,})
    if which in ['total','both']:
        meta_out_rog.update({
            'rog_total': np.float64,
            'com_total_'+latColName: np.float64,
            'com_total_'+lonColName: np.float64,
        })
    df_rog = df.groupby(uidColName)\
                .apply(_user_ROG, meta=meta_out_rog,
                        which=which, df_hw_locs=df_hw_locs)
    return df_rog
                

def _user_ROG(g, which='both', df_hw_locs=None):
    compute = True
    n_pings = g.shape[0]
    coords = g[[latColName,lonColName]].values
    tmp_uid = g[uidColName].iloc[-1]
    # print(g[uidColName])
    out_vals = [n_pings]
    out_cols = ['n_pings']
    if which in ['home','both']:
        if tmp_uid in df_hw_locs.index:
            row_user = df_hw_locs.loc[tmp_uid]
            center_of_mass = np.array([row_user[latColName+'_home'],
                                       row_user[lonColName+'_home']]).reshape((1,2))
            rog = rad_of_gyr(coords, center_of_mass=center_of_mass)
        else:
            center_of_mass = np.array([-999., -999.]).reshape((1,2))
            rog = -1.
        out_vals.extend([rog,
                         center_of_mass[0,0],
                         center_of_mass[0,1],])
        out_cols.extend(['rog_home',
                         'com_home_'+latColName,
                         'com_home_'+lonColName,
                        ])
    if which in ['total', 'both']:
        center_of_mass = np.mean(coords, axis=0, keepdims=True)
        rog = rad_of_gyr(coords, center_of_mass=center_of_mass)
        out_vals.extend([rog,
                         center_of_mass[0,0],
                         center_of_mass[0,1],])
        out_cols.extend(['rog_total',
                         'com_total_'+latColName,
                         'com_total_'+lonColName,
                        ])
    return pd.DataFrame([out_vals], columns=out_cols)


def rad_of_gyr(coords: np.array, center_of_mass=None) -> np.array:
    '''
    Parameters
    ----------
    coords : np.array
        a Nx*2 array of (lat,lon) coordinates.
    center_of_mass : np.array, optional
        A (1,2) np array containing the latitude and longitude of the center
        of mass to be used to compute the ROG (for instance, the user's home).
    
    Returns
    -------
    radius_of_gyrations : float
        The radius of gyration for the selected coords.
    '''
    
    if center_of_mass is None:
        center_of_mass = np.mean(coords, axis=0, keepdims=True)
    rog_sum = (haversine_pairwise(coords, center_of_mass)**2.).sum()
    rog = np.sqrt(rog_sum / max(1,coords.shape[0]))
    
    return rog
    

def totalUserTravelDistance(df_pings, doROG=False, freq='1d'):
    '''
    Computes the total distance traveled (km computed on the straight lines
    between each point) by a user i each `freq` time bin.
    
    Parameters
    ----------
    df_pings : dask.DataFrame
        The dataframe containing at least the :attr:`mobilkit.dask_schemas.uidColName`,
        :attr:`mobilkit.dask_schemas.latColName`, :attr:`mobilkit.dask_schemas.lonColName`
        and :attr:`mobilkit.dask_schemas.dttColName`.
    doROG : bool, optional
        If `True` also computes the ROG on the pings of that day.
    freq : str, optional
        The datetime interval to fllor the `dttColName` to (default one day).
    
    Returns
    -------
    ttd : dask.DataFrame
        The dataframe containing the `user,tBin` index and:
        - `ttd` column with the total traveled distance (in km) for that user on that time bin. 
        - `nPings` the number of pings for that user on that time bin;
        - if `doROG` a column `rog` with the daily ROG using as center of mass the mean point
          of that time bin's coordinates.
    '''
    df_pings['tBin'] = df_pings[dttColName].dt.floor(freq)
    if doROG:
        myApply = lambda g: pd.Series({
                                   'ttd': total_distance_traveled(g[[latColName,lonColName]].values),
                                   'rog': rad_of_gyr(g[[latColName,lonColName]].values),
                                   'nPings': g.shape[0],        
                                })
        myMeta = {'ttd':float, 'rog': float, 'nPings': int}
    else:
        myApply = lambda g: pd.Series({
                                   'ttd': total_distance_traveled(g[[latColName,lonColName]].values),
                                   'nPings': g.shape[0],        
                                })
        myMeta = {'ttd':float, 'nPings': int}
    ttd = df_pings[[uidColName,'tBin',latColName,lonColName]]\
                    .groupby([uidColName,'tBin'])\
                    .apply(myApply, meta=myMeta)
    return ttd


def total_distance_traveled(coords):
    '''
    Parameters
    ----------
    coords : np.array
        a Nx*2 array of (lat,lon) coordinates.
    
    Returns
    -------
    total_distance_traveled : float
        The total distance traveled in the dataframe.
    '''
    tot_dist = 0
    for i, j in zip(coords[:-1,:], coords[1:,:]):
        tot_dist += haversine_pairwise([i],[j])[0][0]
    
    return tot_dist
    
# POIs tools

def convert_df_crs(df,
                   lat_col="lat",
                   lon_col="lng",
                   from_crs="EPSG:4326",
                   to_crs="EPSG:6362",
                   return_gdf=False,
                   ):
    '''
    Parameters
    ----------
    df : pandas.DataFrame
        A dataframe containing the `lat_col` and `lon_col` columns at least.
    lat_col, lon_col : str
        The names of the columns containing the latitude and longitude of the
        points in `df`.
    from_crs, to_crs : str, optional
        The codes of the original and target projections to use.
        If `to_crs` is None no reprojection is done.
    return_gdf : bool, optional
        If `True` returns the newly created gdf otherwise the original df with
        two additional columns telling the projected lat and lon.
        
    Returns
    -------
    df : pd.DataFrame or gpd.GeoDataFrame
        If `return_gdf` the df ported to a geodataframe in the `to_crs` projection.
        Otherwise, the original data frame with two additional columns named `lat_col + '_proj'`
        and `lon_col + '_proj'` containing the original coordinates projected to
        `to_crs`.
    '''
    gdf = gpd.GeoDataFrame(df.copy(),
                           geometry=gpd.points_from_xy(df[lon_col],df[lat_col]),
                           crs=from_crs,
                           )
    if to_crs is not None:
        gdf.to_crs(to_crs, inplace=True)

    if return_gdf:
        return gdf
    else:
        df = df.assign(**{
            lon_col + "_proj": gdf["geometry"].x.values,
            lat_col + "_proj": gdf["geometry"].y.values})
        return df

def distanceHomeUser(g,
                     lon_col=lonColName, lat_col=latColName,
                     h_lon_col="homelon", h_lat_col="homelat"):
    '''
    Parameters
    ----------
    g : pandas.DataFrame
        A dataframe containing at least the `lat_col` and `lon_col` columns with
        the raw points' coordinates and the home coordinates in `homelon` and
        `homelat` columns. **Must contain all the data of one user only.**
    lat_col, lon_col : str
        The names of the columns containing the latitude and longitude of the
        points in `g`.
    h_lat_col, h_lon_col : str
        The names of the columns containing the latitude and longitude of the
        home user `g`.
        
    Returns
    -------
    g : pd.DataFrame
        The original data frame with an additional column named `'home_dist'`
        containing the haversine distance between each point and the home location
        **in kilometers**.
        
    
    Note
    ----
    When determining the home location of a user, please consider that some data providers, like *Cuebiq*, obfuscate/obscure/alter the coordinates of the points falling near the user's home location in order to preserve privacy.

    This means that you cannot locate the precise home of a user with a spatial resolution higher than the one used to obfuscate these data. If you are interested in the census area (or geohash) of the user's home alone and you are using a spatial tessellation with a spatial resolution wider than or equal to the one used to obfuscate the data, then this is of no concern.

    However, tasks such as stop-detection or POI visit rate computation may be affected by the noise added to data in the user's home location area. Please check if your data has such noise added and choose the spatial tessellation according to your use case.
    '''
    home_lat_lon = g[[h_lat_col,h_lon_col]].iloc[:1].values
    tmp_coordinates = g[[lat_col,lon_col]].values
    distances = haversine_pairwise(tmp_coordinates, home_lat_lon)
    g["home_dist"] = distances[:,0]
    return g

def user_dist_cbds(df_hw,
                  cbds_latlon,
                  assign_lat_col='lat_work',
                  assign_lng_col='lng_work',
                  distance_lat_col='lat_home',
                  distance_lng_col='lng_home',
                 ):
    '''
    Computes the distance from the CBD for each user using the cbd which is closest
    to the assign lat lons and computing its distance from distance_lat/lng.
    '''
    
    df_hw['closest_cbd_idx'] = df_hw.apply(lambda r: np.argmin(haversine_pairwise([[r[assign_lat_col], r[assign_lng_col]]],
                                                                         cbds_latlon)[0,:], axis=0)
                                           if not (pd.isna(r[assign_lat_col])
                                                   or pd.isna(r[assign_lng_col])
                                                  ) else None
                                           , axis=1)
    
    df_hw['closest_cbd_dist'] = df_hw.apply(lambda r: haversine_pairwise([[r[distance_lat_col], r[distance_lng_col]]],
                                                                        cbds_latlon[int(r['closest_cbd_idx']),:].reshape(1,-1))[0,0]
                                           if not (pd.isna(r[distance_lat_col])
                                                   or pd.isna(r[distance_lng_col])
                                                   or pd.isna(r['closest_cbd_idx'])
                                                  ) else None
                                           , axis=1)
    return df_hw

def distanceHomeDF(g, **kwargs):
    '''
    Parameters
    ----------
    g : pandas.DataFrame
        A dataframe containing at least the `lat_col` and `lon_col` columns with
        the raw points' coordinates and the home coordinates in `homelon` and
        `homelat` columns of all the users.
    **kwargs
        Such as `lat_col`, `lon_col` will be passed to :attr:`mobilkit.spatial.distanceHomeUser`.
        
    Returns
    -------
    g : pd.DataFrame
        The original data frame with an additional column named `'home_dist'`
        containing the haversine distance between each point and the home location
        of the user of that row **in kilometers**.
    '''
    if g.shape[0] > 0:
        g = g.groupby(uidColName).apply(distanceHomeUser, **kwargs)
    else:
        g["home_dist"] = []
    return g

def compute_poi_index_dist(g, tree_model=None,
                           lon_col="lng_proj", lat_col="lat_proj"):
    '''
    Parameters
    ----------
    g : pandas.DataFrame
        A dataframe containing at least the `lat_col` and `lon_col` columns with
        the raw points' coordinates projected to the same projection of the `tree_model`.
    tree_model : sklearn.neighbors.KDTree
        A KDTree trained on the POIs projected in the same proj of lat on lon points.
        Distance will be computed by the tree.
    lat_col, lon_col : str
        The names of the columns containing the latitude and longitude of the points in `g`.
        By default they match the ones used in :attr:`mobilkit.spatial.compute_poi_visit`.
       
    Returns
    -------
    g : pd.DataFrame
        The original data frame with two additional columns named:
            - `'poi_distance'` the distance of the closest poi found in the tree
                **in KM**;
            - `'_POI_INDEX_'` the 0-based index of the closest tree leaf.
    '''
    if g.shape[0] > 0:
        distances, poi_indexes = tree_model.query(g[[lon_col,lat_col]].values)
        distances = distances[:,0] / 1000. # In Km
        poi_indexes = poi_indexes[:,0]
    else:
        distances = []
        poi_indexes = []
    g["poi_distance"] = distances
    g["_POI_INDEX_"] = poi_indexes
    return g

def filter_to_box(df, minlon, maxlon, minlat, maxlat,
                 lat_col=latColName, lon_col=lonColName):
    '''
    Parameters
    ----------
    df : DataFrame
        A dataframe containing at least the `lat_col` and `lon_col` columns with
        the raw points' coordinates.
    {min,max}{lat,lon} : float
        The min and max values of lat and lon (will keep all coords >= min and <= max).
    lat_col, lon_col : str
        The names of the columns containing the latitude and longitude of the points in `df`.
       
    Returns
    -------
    df : pd.DataFrame
        The original data frame filtered to the points within the box.
    '''
    return df[(df[lat_col].between(minlat, maxlat))
              & (df[lon_col].between(minlon, maxlon))]
    
def compute_poi_visit(df_pings, df_homes, df_POIs,
                      from_crs="EPSG:4326", to_crs="EPSG:6362",
                      min_home_dist_km=.2, visit_time_bin="1H",
                      lat_lon_tol_box=.02):
    '''
    Computes the set of users and number of users visiting a given POI for each
    `visit_time_bin` period of time found in the pings dataframe.
    
    Parameters
    ----------
    df_ping : dask.DataFrame
        A dataframe containing at least the `uid`, `datetime`, `lat` and `lng`
        columns with the raw points' coordinates.
        The coordinates must be given in the `from_crs` projection.
    df_homes : Dataframe
        A `pandas` or `dask` Dataframe with the `uid`, `homelat` and `homelon`
        home coordinates of all the users in the df.
        The coordinates must be given in the `from_crs` projection.
        Note that the three dataframes of pings, homes and POIs **must** feature
        the same initial projection equal to `from_crs`.
    df_POIs : Dataframe
        A `pandas` or `dask` Dataframe with at least the `radius`, `poilat` and
        `poilon` columns stating the radius to be considered in the POI (in km)
        and the POI's coordinates.
        The coordinates must be given in the `from_crs` projection.
    from_crs, to_crs : str
        The codes of the original and target projections to use. Will be used to
        compute planar distances in km using a euclidean distance so use the
        appropriate reference system for your ROI (e.g., use `to_crs='EPSG:6362''`
        for the Mexico area).
        Will be passed to :attr:`mobilkit.spatial.convert_df_crs`.
    min_home_dist : float
        The minimum distance for a point to be from the user's home to be considered
        valid (in km).
    visit_time_bin : str
        The frequency of the time bin to use. Each `datetime` will be floored to this
        time frequency.
    lat_lon_tol_box : float
        The pings will be filtered within the box of the maximum/minimum
        latitude/longitude of the POIs original projection's dataframe.
        This is the margin added around this box to account for pings right outside
        of the POIs' boundaries that may still fall into their radius.
       
    Returns
    -------
    pings_merged_home_poi, results : dask.DataFrame, pd.DataFrame
        - `pings_merged_home_poi` is a view on the `dask.DataFrame` containing, for all
            the points falling within the POIs radius and far enough from users' home:
                - the original pings columns plus their projected coords in `{lat,lng}_proj`;
                - the `home` and 'poi' original and projected (with '_proj' suffix) lat coords;
                - 'poi_distance', '_POI_INDEX_' the distance (in km) and the unique index of the
                    closest POI;
                - all the `df_POIs` columns related to this POI (if common names of columns are
                    found they will be inserted with the `_FROM_POI_TABLE` suffix);
                - 'home_dist' the distance in km from the user's home;
                - 'time_bin' the original datetime floored to `visit_time_bin` freq.
        - `results` is a dataframe containing, for each unique `_POI_INDEX_` and `time_bin` as
            given by `visit_time_bin`:
                - all the `df_POIs` columns related to this POI;
                - `users,num_users` the columns containing the list of the `uid`-s of the users
                    found in that POI and that `time_bin` and their number.
    '''
    # prepare aux DFs
    if type(df_homes) == pd.DataFrame:
        df_homes = dd.from_pandas(df_homes, npartitions=1)
    df_homes = df_homes.repartition(npartitions=1)
    if type(df_POIs) == pd.DataFrame:
        df_POIs = dd.from_pandas(df_POIs, npartitions=1)
    df_POIs = df_POIs.repartition(npartitions=1)
    df_POIs = df_POIs.assign(_POI_INDEX_=da.arange(df_POIs.shape[0].compute(), dtype=np.int64))
    
    # Limiting to box
    df_POIs_pd = df_POIs.compute()
    maxlat, minlat = df_POIs_pd["poilat"].max()+lat_lon_tol_box,\
                        df_POIs_pd["poilat"].min()-lat_lon_tol_box
    maxlon, minlon = df_POIs_pd["poilon"].max()+lat_lon_tol_box,\
                        df_POIs_pd["poilon"].min()-lat_lon_tol_box
    
    # Project the home locations, raw pings and POIs locations to the selected crs
    projected_pings = filter_to_box(df_pings, minlon, maxlon, minlat, maxlat)
    projected_pings = projected_pings.map_partitions(convert_df_crs, from_crs=from_crs, to_crs=to_crs)
    projected_POIs = df_POIs.map_partitions(convert_df_crs, from_crs=from_crs, to_crs=to_crs,
                                           lat_col="poilat", lon_col="poilon",)
    projected_homes = df_homes.map_partitions(convert_df_crs, from_crs=from_crs, to_crs=to_crs,
                                            lat_col="homelat", lon_col="homelon",)
    
    # Prepare the Kdtree model
    tree_model = KDTree(projected_POIs[["poilon_proj","poilat_proj"]].compute(), leaf_size=20, metric="euclidean")

    # Check all is good
    assert 1 == projected_POIs._POI_INDEX_.value_counts().compute().max()
    # Check that we do not have double POIs
    distances, poi_indexes = tree_model.query(projected_POIs[["poilon_proj","poilat_proj"]].values)
    assert np.all(poi_indexes[:,0] == np.arange(len(poi_indexes), dtype=int))
    
    # Add the home location
    pings_merged_home = projected_pings.merge(projected_homes, on="uid", how="inner")
    
    pings_merged_home_poiID = pings_merged_home.map_partitions(compute_poi_index_dist,
                                                               tree_model=tree_model)
    pings_merged_home_poi = pings_merged_home_poiID.merge(projected_POIs,
                                                          on="_POI_INDEX_",
                                                          how="inner",
                                                          suffixes=("","_FROM_POI_TABLE"))
    
    # Filter on poi dist radius, then home on remaining
    pings_merged_home_poi = pings_merged_home_poi[pings_merged_home_poi["poi_distance"]
                                                  <= pings_merged_home_poi["radius"]]
    pings_merged_home_poi = pings_merged_home_poi.map_partitions(distanceHomeDF)
    pings_merged_home_poi = pings_merged_home_poi[pings_merged_home_poi["home_dist"] > min_home_dist_km]
    
    # Round datetime to time bin
    pings_merged_home_poi = pings_merged_home_poi.assign(time_bin=
                                                         pings_merged_home_poi["datetime"]\
                                                                 .dt.floor(visit_time_bin))
    
    # Count the users per poi and time bin
    def tmp_foo(g):
        usrs = list(g["uid"].unique())
        n_us = len(usrs)    
        return pd.DataFrame([[usrs, n_us]], columns=["users", "num_users"])

    results = pings_merged_home_poi.groupby(["_POI_INDEX_","time_bin"]).apply(tmp_foo,
                                                                              meta={
                                                                                  "users": object,
                                                                                  "num_users": int,
                                                                              }).compute()
    # merge with df_POIs_pd
    results = results.join(df_POIs_pd.set_index("_POI_INDEX_"), how="inner").reset_index()
    
    return pings_merged_home_poi, results
    

def compute_population_density(df, **kwargs):
    '''
    Parameters
    ----------
    df : dask.DataFrame
        A dataframe as returned by :attr:`mobilkit.temporal.filter_daynight_time` with
        at least the `date,daytime,uid,lat,lng` columns containing the date rounded to
        day, a bool stating if the point is in daytime or nightime, the user id and the
        coordinates of the point.
    **kwargs
        Will be passed to :attr:`mobilkit.spatial.meanshift`.
    
    Returns
    -------
    df : pandas.DataFrame
        A dataframe with a multi index of `date,daytime,uid` and as columns the `lat`
        and `lng` coordinates of the mean shift location of the user on that part of
        the day on that date.
    '''
    res = df.groupby(["date","daytime","uid"]).apply(meanshift, **kwargs).compute()
    res = pd.DataFrame(res, columns=["pppointtt"])
    res = res.assign(**{latColName: res['pppointtt'].apply(lambda v: v[1]),
                         lonColName: res['pppointtt'].apply(lambda v: v[0])})
    res = res[[latColName, lonColName]].copy()
    return res


def density_map(latitudes, longitudes, center, bins, radius):
    '''
    Parameters
    ----------
    latitudes, longitudes : array-like
        The arrays containing the latitude and longitude coordinates of each user's
        location.
    center : tuple-like
        The (lat, lon) of the center where to compute the population density around.
    bins : int
        The number of bins to use horizontally and vertically in the region around the
        center.
    radius : float
        The space to consider above, below, left and right of the center (same unity
        of the center).
    
    Returns
    -------
    density : np.array
        The 2d histogram of the population.
    
    '''
    # Center the map around the provided center coordinates
    histogram_range = [
        [center[1] - radius, center[1] + radius],
        [center[0] - radius, center[0] + radius]
    ]
    res = np.histogram2d(longitudes, latitudes, bins=bins, range=histogram_range)
    return res

from pandas import IndexSlice

def stack_density_map(df, dates, center, daytime=True, bins=200, radius=1):
    '''
    Parameters
    ----------
    df : pd.DataFrame
        A dataframe as returned by :attr:`mobilkit.spatial.compute_population_density`
        with a multi index of `date,daytime,uid` and as columns the `lat` and `lng`
        coordinates of the mean shift location of the user on that part of
        the day on that date.
    dates : list pof datetime
        A list of dates when to compute the density.
    center : tuple-like
        The (lat, lon) of the center where to compute the population density around.
    daytime : bool
        Whether to compute the density on the daytime or nightime part of selected
        dates.
    bins : int
        The number of bins to use horizontally and vertically in the region around the
        center.
    radius : float
        The space to consider above, below, left and right of the center (same unity
        of the center).
    
    Returns
    -------
    maps, results : np.array, tuple
        `maps` is the tensor of shape `(len(dates),bins,bins)` storing for each date the x-y
        density map as computed by :attr:`mobilkit.spatial.density_map`.
        `results` stores the x and y bins.
    '''
    maps = np.zeros((len(dates),bins,bins))
    slicer = IndexSlice
    for iii, date in enumerate(dates):
        tmp_rows = df.loc[slicer[date,daytime,:],:]
        tmp_res = density_map(tmp_rows[latColName], tmp_rows[lonColName],
                              center=center, bins=bins, radius=radius)
        maps[iii,:,:] = tmp_res[0].copy()
    return maps, tmp_res[1:]

import numpy.ma as ma
def stats_density_map(df, dates, center, daytime=True, bins=200, radius=1, clip_std=1e-4):
    '''
    Parameters
    ----------
    df : pd.DataFrame
        A dataframe as returned by :attr:`mobilkit.spatial.compute_population_density`
        with a multi index of `date,daytime,uid` and as columns the `lat` and `lng`
        coordinates of the mean shift location of the user on that part of
        the day on that date.
    dates : list pof datetime
        A list of dates when to compute the density.
    center : tuple-like
        The (lat, lon) of the center where to compute the population density around.
    daytime : bool
        Whether to compute the density on the daytime or nightime part of selected
        dates.
    bins : int
        The number of bins to use horizontally and vertically in the region around the
        center.
    radius : float
        The space to consider above, below, left and right of the center (same unity
        of the center).
    clip_std : float
        Pixels with a 0 or `nan` std will be clipped to this value when computing the
        z-score. The same pixels will be set to -1 on output.
    
    Returns
    -------
    results : dict
        A dictionary containing teh key-values:
            - `stack` the tensor of shape `(len(dates),bins,bins)` storing for each date
                the x-y density map as computed by :attr:`mobilkit.spatial.density_map`.
            - `avg`, `std` the average and standard deviation population density with
                shape `(1,bins,bins)`.
            - `x_bins`, `y_bins` the bins of the 2d histogram as produced by
                :attr:`mobilkit.spatial.density_map`.
    '''
    tensor_maps, bins = stack_density_map(df, dates, center, daytime=daytime, bins=bins, radius=radius)
    maps_avg = tensor_maps.mean(axis=0, keepdims=True)
    maps_std = tensor_maps.std(axis=0, keepdims=True)
    mask = np.where(np.logical_or(np.isnan(maps_std), maps_std==.0))
    maps_std = np.where(np.logical_or(np.isnan(maps_std), maps_std==.0),
                        np.ones_like(maps_std)*clip_std, maps_std)
    maps_zsc = np.zeros_like(tensor_maps)
    maps_zsc = (tensor_maps - maps_avg) / maps_std
    
    maps_std[mask] = -1

    return {"stack": tensor_maps, "zsc": maps_zsc,
            "avg": maps_avg, "std": maps_std,
            "x_bins": bins[0], "y_bins": bins[1]}


def box2poly(box):
    '''
    [min_lon, min_lat, max_lon, max_lat]
    '''
    return shapely.geometry.Polygon([[box[0],box[3]], [box[2],box[3]], [box[2],box[1]],
                                [box[0],box[1]], [box[0],box[3]]])

def makeVoronoi(gdf):
    x = gdf["geometry"].x.values
    y = gdf["geometry"].y.values

    coords = np.vstack((x, y)).T
    vor = Voronoi(coords)

    lines = [shapely.geometry.LineString(vor.vertices[line]) for line in 
        vor.ridge_vertices if -1 not in line]
    polys = shapely.ops.polygonize(lines)
    voronois = gpd.GeoDataFrame(geometry=gpd.GeoSeries(polys), crs=gdf.crs)

    tmp_box = box2poly(gdf.unary_union.bounds)
    tmp_gdf = gpd.GeoDataFrame(["box"], geometry=[tmp_box], crs=gdf.crs)
    voronois = gpd.overlay(voronois, tmp_gdf)
    
    return voronois


# Copyright (C) MindEarth <enrico.ubaldi@mindearth.org> @ Mindearth 2020-2021
# 
# This file is part of mobilkit.
#
# mobilkit is distributed under the MIT license.

### import key libraries
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import glob, os
from datetime import datetime as dt
from datetime import timedelta
from datetime import timezone
import pytz
from copy import copy, deepcopy

from matplotlib.colors import LogNorm
import matplotlib.colors as mcolors
from sklearn.metrics import r2_score
from sklearn.linear_model import LinearRegression
import seaborn as sns

import geopandas as gpd
import pyproj

from mobilkit.dask_schemas import (
    accColName,
    lonColName,
    latColName,
    uidColName,
    utcColName,
    dttColName,
    zidColName,
)

def compareLinePlot(x_scatter, x_line, y, data,
                    xlim=None,
                    ylim=None,
                    xlabel=None,
                    ylabel=None,
                    doScatter=True,
                    doLine=True,
                    scatterkws={},
                    lineplotkws={},
                    figsize=(7,4),
                    ax=None,
                   ):
    '''
    Compares a scattered data with its line estimated.
    
    Parameters
    ----------
    x_scatter, x_line, y : str
        The columns to use for x in the scatter and line plot (in the line plot
        you might want to use a binned version of the x) and as y.
    data : pd.Dataframe
        The dataframe to use.
    xlim, ylim :  tuple, optional
        The limits to put in the x and y axis.
    xlabel, ylabel : str, optional
        The x and y axis labels
    scatterkws :  dict, optional
        The keywords to pass to `seaborn.scatterplot`.
        By default thay are:
        `{'alpha':.075}`
    lineplotkws :  dict, optional
        The keywords to pass to `seaborn.lineplot`.
        By default thay are:
        `{'color':'C3',
          'estimator':lambda g: np.percentile(g, 50),
          'n_boot':200}`
    figsize : tuple, optional
        The figure size in inches.
    Returns
    -------
    fig, ax : tuple
        The figure and axes handle.
    '''
    if ax is None:
        fig, ax = plt.subplots(1,1,figsize=figsize)
    else:
        fig = ax.figure

    scakws = {
        'alpha':.075
    }
    linkws = {
        'color':'C3',
        'estimator':lambda g: np.percentile(g, 50),
        'n_boot':200
    }
    
    scakws.update(**scatterkws)
    linkws.update(**lineplotkws)
    
    if doScatter:
         ax = sns.scatterplot(
            data=data,
            x=x_scatter,
            y=y,
            ax=ax,
            **scakws
        )
    
    if doLine:
        ax = sns.lineplot(
            data=data,
            x=x_line,
            y=y,
            ax=ax,
            **linkws
        )

    if xlim:
        ax.set_xlim(*xlim)
    if ylim:
        ax.set_ylim(*ylim)
        
    if xlabel:
        ax.set_xlabel(xlabel)
    else:
        ax.set_xlabel(x_scatter)
    if ylabel:
        ax.set_ylabel(ylabel)
    else:
        ax.set_ylabel(y)
    
    return fig, ax

def visualize_simpleplot(df):
    import contextily as ctx
    gdf = gpd.GeoDataFrame(df, 
                           geometry=gpd.points_from_xy(df[lonColName],
                                                       df[latColName]),
                           crs= {"init": "epsg:4326"}).to_crs(epsg=3857)
    fig,ax = plt.subplots(figsize=(15,10))
    xmin, ymin, xmax, ymax = gdf.total_bounds
    basemap, extent = ctx.bounds2img(xmin, ymin, xmax, ymax, 
                                     source=ctx.providers.CartoDB.Voyager)
    gdf.plot(ax=ax, zorder=2)
    ax.imshow(basemap, extent=extent, zorder=1)
    plt.show()

def visualize_boundarymap(boundary):
    import contextily as ctx
    fig,ax = plt.subplots(figsize=(15,10))
    x1, y1 = (boundary[0], boundary[1])
    x2, y2 = (boundary[2], boundary[3])
    basemap, extent = ctx.bounds2img(x1, y1, x2, y2, ll=True,
                                     source=ctx.providers.CartoDB.Voyager)
    ax.imshow(basemap, extent=extent)
    plt.show()
    
    
def plot_density_map(latitudes, longitudes, center, bins, radius, ax=None, annotations=None):
    '''
    Parameters
    ----------
    latitudes, longitudes : array-like
        Array contaning the lat and lon coordinates of each user on a selected day.
    center : tuple-like
        The (lat, lon) of the center where to compute the population density around.
    bins : int
        The number of bins to use horizontally and vertically in the region around the
        center.
    radius : float
        The space to consider above, below, left and right of the center (same unity
        of the center).
    ax : matplotlib.axes
        The axes to use. If `None` a new figure will be created.
    annotations : dict
        A dictionary of annotations to be put on the map in the form of
        `{"Text": {kwargs to ax.annotate} }`. Will be used as `ax.annotate(key, **value)`.
    
    Returns
    -------
    res, ax
        The putput of `ax.hist2d` and the axis itself.
    '''
    
    cmap = copy(plt.cm.jet)
    cmap.set_bad((0,0,0))  # Fill background with black

    # Center the map around the provided center coordinates
    histogram_range = [
        [center[1] - radius, center[1] + radius],
        [center[0] - radius, center[0] + radius]
    ]
    
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(5,5))
        ax.set_aspect("equal")
        
    res = ax.hist2d(longitudes, latitudes, bins=bins, norm=LogNorm(),
                       cmap=cmap, range=histogram_range)
        
    if annotations:
        for name, annot in annotations.items():
            ax.annotate(name, **annot)
    ax.axis('equal')
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.grid('off')
    ax.axis('off')
    return res, ax
    
import matplotlib.colors as mcolors
import mpl_toolkits
import matplotlib.gridspec as gridspec
def shori_density_map(data, xbins, ybins, ax=None, annotations=None, vmin=-2, vmax=2):
    '''
    Parameters
    ----------
    data : array-like
        Array contaning the raster of the population density to be plot.
    xbins, ybins : array-like
        The bins used to construct the raster. Will be used to limit the plot area.
    ax : matplotlib.axes
        The axes to use. If `None` a new figure will be created.
    annotations : dict
        A dictionary of annotations to be put on the map in the form of
        `{"Text": {kwargs to ax.annotate} }`. Will be used as `ax.annotate(key, **value)`.
    vmin, vmax : float
        The values to be passed to the colormap.
    
    Returns
    -------
    res
        The output of `ax.imshow`.
    '''
    
    divnorm = mcolors.SymLogNorm(linthresh=1.0, linscale=1.0, vmin=vmin, vmax=vmax)
    cmap = copy(plt.cm.jet)
    diff = np.ma.masked_where(data == 0, data)
    cmap.set_bad(color='black')  # Fill background with black
    
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(5,5))
        # ax.set_aspect("equal")
    else:
        fig = ax.figure
        
    res = ax.imshow(diff.T, cmap=cmap,
                    extent=[xbins[0],xbins[-1],ybins[0],ybins[-1]], 
                    norm=divnorm, origin="lower")
    # divider = mpl_toolkits.axes_grid1.make_axes_locatable(ax)
    # cax = divider.append_axes('right', '5%', pad='3%')
    # fig.colorbar(res, cax=cax)
    
    if annotations is not None:
        for name, kwar in annotations.items():
            ax.annotate(name, **kwar)
    # ax.axis('equal')
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.grid('off')
    ax.axis('off')
    
    return res


def plot_pop(df, title, empiric_pop="POBTOT", data_pop="POP_HFLB", alpha=.1, verbose=True):
    '''
    Plot the scatter-plot between empiric_pop and data_pop columns.
    
    Parameters
    ----------
    df : pandas.DataFrame
        As dataframe containing the two columns selected.
    '''

    plt.scatter(df[empiric_pop], df[data_pop], alpha=alpha)
    plt.loglog();

    plt.xlim(1, df[empiric_pop].max()*2.)
    plt.ylim(1, df[data_pop].max()*2.)

    lr = LinearRegression(fit_intercept=False)

    valid = df[(~df[empiric_pop].isna())
                          & (~df[data_pop].isna())]

    X = valid[empiric_pop].values
    Y = valid[data_pop].values

    idxs = np.logical_and(X>0,Y>0)
    X = X[idxs]
    Y = Y[idxs]

    lr.fit(X.reshape(-1,1),
           Y.reshape(-1,1))

    X_pred = np.array(sorted(X)).reshape(-1,1)
    Y_pred = lr.predict(X_pred)

    import statsmodels.api as sm
    from scipy import stats

    X2 = sm.add_constant(np.log10(X))
    est = sm.OLS(np.log10(Y), X2)
    est2 = est.fit()

    if verbose:
        print(est2.summary())

    plt.plot(X_pred, Y_pred, "--C1", label="Slope: %.02f"
             % (est2.params[1]))
    
    plt.title(title)

    plt.legend()
    plt.xlabel(empiric_pop + title)
    plt.ylabel(data_pop)

    return est2
